// ESP32 Running 2 Cores
// January February 2019
// Written by Theron Wierenga
// theron@wierengaconsulting.com

#include <WiFi.h>
#include <HTTPClient.h>
#include "RTClib.h"
#include "Wire.h"
#include "time.h"
#include "SparkFunBME280.h"
#include <LiquidCrystal_I2C.h>
#include "driver/i2c.h"

#define SEG_A 14
#define SEG_B 15
#define SEG_C 32
#define SEG_D 25
#define SEG_E 26
#define SEG_F 13
#define SEG_G 27
#define DP 33
#define EN 16
#define INT_LED 17
#define ADD0 0
#define ADD1 2
#define ADD2 4
#define MODE0 34    // For which SSID to use
#define MODE1 35    // For which weather report

TaskHandle_t Task1;

// Your SSID name and passwords mujst be entered here
const char* ssid       = "SSID1";
const char* password   = "password1";

const char* ssid2       = "SSID2";
const char* password2   = "password2";

const char* ntpServer = "pool.ntp.org";
const long  gmtOffset_sec = -3600 * 5;   // Eastern Standard Time
const int   daylightOffset_sec = 3600;

// Create Real Time Clock instance
RTC_DS3231 rtc;
DateTime now_time;
BME280 mySensor;
HTTPClient http;
LiquidCrystal_I2C lcd(0x3F,20,4);

// Global variables for time
byte bsecond, stemp, bminute, mtemp, bhour, htemp, ampm, bdayOfWeek, bdayOfMonth, bmonth, b1, b2, b3, b4, last_sec = 0;
byte adj_bsecond, adj_bminute, adj_bhour, adj_bdayOfMonth, adj_bmonth;
byte h10, h1, m10, m1, s10, s1;
unsigned int iyear, adj_iyear;
String payload, time_string, weather, wtemp;
time_t now;
struct tm timeinfo;
  // int tm_sec;
  // int tm_min;
  // int tm_hour;
  // int tm_mday;
  // int tm_mon;
  // int tm_year;
  // int tm_wday;
  // int tm_yday;
  // int tm_isdst;

String daysOfTheWeek[7] = {"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"};
String months[12] = {"January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"};
String tags[13] = {"main", "description", "temp", "pressure", "humidity", "temp_min", "temp_max",
                   "speed", "deg", "gust", "sunrise", "sunset", "name"};
int num_flag[13] = {0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0}; 
String Desc, Name, Deg, Temp, Humid, Speed;
// 7 Segment patterns  DP, A, B, C, D, E, F, G    7 - 0
byte num[] = {B10000001, B11001111, B10010010, B10000110, B11001100, B10100100, B11100000, B10001111, B10000000, B10001100};
byte H = B11001000, C = B10110001, blank = B11111111, F = B10111000, dec_point = B01111111;
byte mode0, mode1;
int del = 2000, current_int = 0, code, lcd_delay = 1;
bool debug = false, debug_min = true;
double currenttemp = 12.3, humidity = 28.9, farentemp = 17.6;
String home_location = "4175331", home_location2 = "5004005";


void Task1code( void * parameter) 
{
   // Task here must never end or return
   while(true)
   {
      // Core 0 just scans LEDs

      switch (last_sec)
      {
         case 1: case 2: case 3: case 4: case 5: date_one_scan(); break;
         case 31: case 32: case 33: case 34: case 35: temp_hum_one_scan(); break;
         default: time_one_scan();
      }
   }
}

void setup() 
{  
    Serial.begin(115200);
    delay(3000);
    if (debug_min)
    {
        Serial.print("setup() running on core ");
        Serial.println(xPortGetCoreID());
       
        Serial.print("SDA on pin ");
        Serial.println(SDA);
        Serial.print("SCL on pin ");
        Serial.println(SCL);
    }
    pinMode(SEG_A, OUTPUT);
    pinMode(SEG_B, OUTPUT);
    pinMode(SEG_C, OUTPUT);
    pinMode(SEG_D, OUTPUT);
    pinMode(SEG_E, OUTPUT);
    pinMode(SEG_F, OUTPUT);
    pinMode(SEG_G, OUTPUT);
    pinMode(DP, OUTPUT);
    pinMode(EN, OUTPUT);
    pinMode(INT_LED, OUTPUT);
    pinMode(ADD0, OUTPUT);
    pinMode(ADD1, OUTPUT);
    pinMode(ADD2, OUTPUT);
    pinMode(MODE0, INPUT);
    pinMode(MODE1, INPUT);

    payload.reserve(1000);
    weather.reserve(200);

    if (! rtc.begin())
    {
        Serial.println("Couldn't find RTC");
        while (1);
    }
  
    // Connect to WiFi
    mode0 = digitalRead(MODE0);
    if (debug)
    {
        Serial.print("mode0 = ");
        Serial.println(mode0);
    }
    Serial.printf("Connecting to %s ", ssid);
    if (mode0 == 0)
    {
        WiFi.begin(ssid, password);
    }
    else
    {
        WiFi.begin(ssid2, password2);
    }
    while (WiFi.status() != WL_CONNECTED) {
        delay(500);
        Serial.print(".");
    }
    Serial.println(" CONNECTED");
    digitalWrite(INT_LED, HIGH);
    
    // Init and get the time
    configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);
    getLocalTime(&timeinfo);
    if (debug_min)
    {
        Serial.print("In Setup - Time from Internet  ");
        Serial.print(timeinfo.tm_hour);
        Serial.print(" : ");
        Serial.print(timeinfo.tm_min);
        Serial.print(" : ");
        Serial.println(timeinfo.tm_sec);
    }
    iyear = timeinfo.tm_year;
    iyear = iyear % 100;
    iyear = iyear + 2000;
    bmonth = timeinfo.tm_mon;
    bdayOfMonth = timeinfo.tm_mday;
    bhour = timeinfo.tm_hour;
    bminute = timeinfo.tm_min;
    bsecond = timeinfo.tm_sec;
    rtc.adjust(DateTime(iyear, bmonth, bdayOfMonth, bhour, bminute, bsecond));
  //  rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));
  //  rtc.adjust(DateTime(2019, 1, 25, 20, 33, 30));
    delay(10);
    now_time = rtc.now();
    bhour = now_time.hour();
    bminute = now_time.minute();
    last_sec = now_time.second();
    if (debug_min)
    {
        Serial.print("In Setup - Time from RTC      ");
        Serial.print(bhour);
        Serial.print(" : ");
        Serial.print(bminute);
        Serial.print(" : ");
        Serial.println(last_sec);
    }
    mySensor.settings.commInterface = I2C_MODE;
    mySensor.settings.I2CAddress = 0x76;
    mySensor.settings.runMode = 3; //Normal mode
    mySensor.settings.tStandby = 0;
    mySensor.settings.filter = 0;
    mySensor.settings.tempOverSample = 1;
    mySensor.settings.pressOverSample = 1;
    mySensor.settings.humidOverSample = 1;
    mySensor.begin();
    mySensor.calibration.dig_T1;
    mySensor.calibration.dig_T2;
    mySensor.calibration.dig_T3;
    mySensor.calibration.dig_P1;
    mySensor.calibration.dig_P2;
    mySensor.calibration.dig_P3;
    mySensor.calibration.dig_P4;
    mySensor.calibration.dig_P5;
    mySensor.calibration.dig_P6;
    mySensor.calibration.dig_P7;
    mySensor.calibration.dig_P8;
    mySensor.calibration.dig_P9;
    mySensor.calibration.dig_H1;
    mySensor.calibration.dig_H2;
    mySensor.calibration.dig_H3;
    mySensor.calibration.dig_H4;
    mySensor.calibration.dig_H5;
    mySensor.calibration.dig_H6;

    // initialize the lcd 
    lcd.init();
    // Print a message to the LCD.
    lcd.backlight();
    lcd.setCursor(0,0);
    lcd.print("Starting");
   
    xTaskCreatePinnedToCore(
      Task1code, // Function to implement the task
      "Task1",   // Name of the task
      2000,     // Stack size in words
      NULL,      // Task input parameter
      1,         // Priority of the task
      &Task1,    // Task handle
      0);        // Core where the task should run
   delay(500);
}

void loop() 
{
    if (WiFi.status() == WL_CONNECTED)
    {
         while (now_time.second() == last_sec)
         {
            now_time = rtc.now();
         }   
         last_sec = now_time.second();
         bminute = now_time.minute();
         bhour = now_time.hour();
         bdayOfMonth = now_time.day();
         bmonth = now_time.month();
         iyear = now_time.year();
         if (debug)
         {
             Serial.print(bhour);
             Serial.print(":");
             Serial.print(bminute);
             Serial.print(":");
             Serial.println(last_sec);
         }
         
         ampm = 0;
         // Check if it's 12 PM
         if (bhour == 12) ampm = 1;
        
         // Check if 12 AM
         if (bhour == 0) bhour = 12;
        
         // Check if it's PM
         if (bhour > 12)
         {
           bhour -= 12;
           ampm = 1;
         }
          
         if (last_sec == 0)                 // Read time and date from Interent and write to RTC
         {
            getLocalTime(&timeinfo);
            adj_iyear = timeinfo.tm_year;
            adj_iyear = adj_iyear % 100;
            adj_iyear = adj_iyear + 2000;
            adj_bmonth = timeinfo.tm_mon;
            adj_bdayOfMonth = timeinfo.tm_mday;
            adj_bhour = timeinfo.tm_hour;
            adj_bminute = timeinfo.tm_min;
            adj_bsecond = timeinfo.tm_sec;
            rtc.adjust(DateTime(adj_iyear, adj_bmonth, adj_bdayOfMonth, adj_bhour, adj_bminute, adj_bsecond));
            delay(10);
            if (debug)
            {
                Serial.print("\nReset time from core ");
                Serial.print(xPortGetCoreID());
                Serial.print(" - ");
                Serial.print(timeinfo.tm_hour);
                Serial.print(":");
                Serial.print(timeinfo.tm_min);
                Serial.print(":");
                Serial.println(timeinfo.tm_sec);
            }
         }

         if (last_sec == 10)              // Read the temperature and humidity from the BME280 and store it
         {
           // Read temperature and humidity
        //   ifaren = int(mySensor.readTempF() * 10 + 2.5);
        //   ihum = int(mySensor.readFloatHumidity() * 10 + 0.5);
           currenttemp = mySensor.readTempF();
           humidity = mySensor.readFloatHumidity();
           if (debug)
           {
               Serial.print("  Temp. = ");
               Serial.print(currenttemp);
               Serial.print("  Hum. = ");
               Serial.println(humidity);
           }
           
           currenttemp = float(int(currenttemp * 10.0 + 0.5)) / 10.0;
           humidity = float(int(humidity * 10.0 + 0.5)) / 10.0;

           if(debug)
           {
               Serial.println(String(currenttemp));
               Serial.println(String(humidity));
           }
         }

         if (last_sec == 20)              // Read weather information from the Interent and store it
         {
            mode1 = digitalRead(MODE1);
            if (debug)
            {
                Serial.print("MODE1 = ");
                Serial.println(mode1);
            }
            if (mode1 == 0)
            {
              code = getWeatherFromInternet(home_location, "Home");
            }
            else
            {
              code = getWeatherFromInternet(home_location2, "Home");
            }
         }

         if (last_sec == 30)            // Write the weather information to the LCD display
         {
          if (debug)
          {
            Serial.print("httpCode = ");
            Serial.println(code);
          }
         //   if (code == 200)
            {
                lcd.setCursor(0, 0);
                while (Name.length() < 15) Name += " ";
                if (bhour < 10) Name+= " ";
                Name += String(bhour) + ":";
                if (bminute < 10) Name += "0";
                Name += String(bminute);
                lcd.print(Name);
                delay(lcd_delay);
                lcd.setCursor(0,1);
                delay(lcd_delay);
                Desc.setCharAt(0, Desc.charAt(0) - 0x20); // Make it a capital
                while (Desc.length() < 20) Desc += " ";
                lcd.print(Desc);
                delay(lcd_delay);
                lcd.setCursor(0,2);
                delay(lcd_delay);
                wtemp = Temp + "F  " + Humid + "%H";
                while (wtemp.length() < 20) wtemp += " ";
                lcd.print(wtemp);
                delay(lcd_delay);
                lcd.setCursor(0,3);
                delay(lcd_delay);
                wtemp = "Wind " + Speed + " from " + Deg;
                while (wtemp.length() < 20) wtemp += " ";
                lcd.print(wtemp);
            }
         }
    }
    else
    {
        // Reconnect if signal lost
        digitalWrite(INT_LED, LOW);
        Serial.println("WiFI lost, restarting...");
        WiFi.disconnect(true);
        mode0 = digitalRead(MODE0);
        if (debug)
        {
            Serial.print("MODE0 = ");
            Serial.println(mode0);
        }
        if (mode0 == 0)
        {
            WiFi.begin(ssid, password);
        }
        else
        {
            WiFi.begin(ssid2, password2);
        }
        while (WiFi.status() != WL_CONNECTED)
        {
            delay(2500);
            Serial.print(".");
        }
        digitalWrite(INT_LED, HIGH);
    }
}

void time_one_scan()
{
  // One scan of digits
  if (bhour / 10 != 0)          // Suppress leading zero
  {
      digitalWrite(EN, HIGH);   // disable outputs
      set_segs(num[bhour / 10]);
      set_digit(7);
      digitalWrite(EN, LOW);    // enable outputs
      delayMicroseconds(del);
  }

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[bhour % 10] - 128);
  set_digit(6);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[bminute / 10] - 128);
  set_digit(5);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[bminute % 10]);
  set_digit(4);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  // Skip digit 3

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[last_sec / 10]);
  set_digit(2);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[last_sec % 10]);
  set_digit(1);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  // Skip digit 0

}

void date_one_scan()
{
  if (bmonth / 10 != 0)          // Suppress leading zero
  {
    digitalWrite(EN, HIGH);   // disable outputs
    set_segs(num[(bmonth + 1) / 10]);
    set_digit(7);
    digitalWrite(EN, LOW);    // enable outputs
    delayMicroseconds(del);
  }

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[(bmonth + 1) % 10]);
  set_digit(6);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  if (bdayOfMonth / 10 != 0)
  {
    digitalWrite(EN, HIGH);   // disable outputs
    set_segs(num[bdayOfMonth / 10]);
    set_digit(5);
    digitalWrite(EN, LOW);    // enable outputs
    delayMicroseconds(del);
  }

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[bdayOfMonth % 10]);
  set_digit(4);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[iyear / 1000]);
  set_digit(3);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[(iyear / 100) % 10]);
  set_digit(2);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[(iyear / 10) % 10]);
  set_digit(1);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[iyear % 10]);
  set_digit(0);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);
}

void temp_hum_one_scan()
{
  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[int(currenttemp) / 10]);
  set_digit(7);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[int(currenttemp) % 10] - 128);
  set_digit(6);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[int(currenttemp * 10.0) % 10]);
  set_digit(5);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(F);
  set_digit(4);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[int(humidity) / 10]);
  set_digit(3);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[int(humidity) % 10] - 128);
  set_digit(2);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(num[int(humidity * 10.0) % 10]);
  set_digit(1);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);

  digitalWrite(EN, HIGH);   // disable outputs
  set_segs(H);
  set_digit(0);
  digitalWrite(EN, LOW);    // enable outputs
  delayMicroseconds(del);
}

void set_segs(byte index)
{
  // if bit = 0, segment is on
  // byte num[] - {B10000001, B11001111, B10010010, B10000110, B11001100, B10100100, B11100000, B10001111, B10000000, B10001100};
 
  digitalWrite(DP, bitRead(index, 7));
  digitalWrite(SEG_A, bitRead(index, 6));
  digitalWrite(SEG_B, bitRead(index, 5));
  digitalWrite(SEG_C, bitRead(index, 4));
  digitalWrite(SEG_D, bitRead(index, 3));
  digitalWrite(SEG_E, bitRead(index, 2));
  digitalWrite(SEG_F, bitRead(index, 1));
  digitalWrite(SEG_G, bitRead(index, 0));
}

void set_digit(byte index)
{
  switch(index)
  {
    case 0: digitalWrite(ADD2, 0);
            digitalWrite(ADD1, 0);
            digitalWrite(ADD0, 0);
            break;
    case 1: digitalWrite(ADD2, 0);
            digitalWrite(ADD1, 0);
            digitalWrite(ADD0, 1);
            break;
    case 2: digitalWrite(ADD2, 0);
            digitalWrite(ADD1, 1);
            digitalWrite(ADD0, 0);
            break;
    case 3: digitalWrite(ADD2, 0);
            digitalWrite(ADD1, 1);
            digitalWrite(ADD0, 1);
            break;
    case 4: digitalWrite(ADD2, 1);
            digitalWrite(ADD1, 0);
            digitalWrite(ADD0, 0);
            break;
    case 5: digitalWrite(ADD2, 1);
            digitalWrite(ADD1, 0);
            digitalWrite(ADD0, 1);
            break;
    case 6: digitalWrite(ADD2, 1);
            digitalWrite(ADD1, 1);
            digitalWrite(ADD0, 0);
            break;
    case 7: digitalWrite(ADD2, 1);
            digitalWrite(ADD1, 1);
            digitalWrite(ADD0, 1);
            break;
  }
}

int getWeatherFromInternet(String location, String isitHome)
{
    const String endpoint = "http://api.openweathermap.org/data/2.5/weather?id=" + location + "&units=imperial&APPID=";
    const String key =      "22759a049ab498051a6d4c7343bd2d08";

    http.begin(endpoint + key); //Specify the URL
    int httpCode = http.GET();  //Make the request

    if (httpCode > 0)
    {
        // Check for the returning code
        payload = http.getString();
        if (debug) 
        {
            Serial.println(payload);
            Serial.print(payload.length());
            Serial.println(" bytes");
        }
        weather = parse_it(payload, isitHome);
     }
     else 
     {
          if (debug) Serial.println("Error on HTTP request");
     }
     
     http.end(); //Free the resources
     return httpCode;
}

String parse_it(String tag, String isitHome)
{
    int tags_i = 0, index = 0, val1, val2, last_val;
    String tempx = "", last_temp = "";
    String ret_string = isitHome;
    ret_string += ",";
    while (tags_i < 13)
    {
        val1 = tag.indexOf(tags[tags_i], 0);
        if (val1 < 0)
        {
            tags_i += 1;
            continue;
        }
        tempx = tag.substring(val1, val1 + tags[tags_i].length());
        if (debug) Serial.println(tags[tags_i]);
        if (tempx == tags[tags_i])
        {
            if (num_flag[tags_i] == 1)    // It's a numeric value
            {
                val1 = val1 + tags[tags_i].length() + 2;
                index = val1;
                val2 = tag.indexOf("\"", index) - 1;
                tempx = tag.substring(val1, val2);
                if (tempx.endsWith("}"))
                {
                  tempx.remove(tempx.length() - 1, 1);
                }     
                if (debug) Serial.print("            ");
                if (debug) Serial.println(tempx);
                if ((tags[tags_i] == "temp") || (tags[tags_i] == "humidity") || (tags[tags_i] == "speed")|| (tags[tags_i] == "deg"))
                {
                    ret_string += ",";
                    ret_string += tempx;
                    if (tags[tags_i] == "temp") Temp = tempx;
                    if (tags[tags_i] == "humidity") Humid = tempx;
                    if (tags[tags_i] == "speed") Speed = tempx;
                    if (tags[tags_i] == "deg") 
                    {
                      // Deg = tempx;
                      float wf = wtemp.toFloat();
                      if (wf < 22.5) Deg = "N";
                      if ((wf >= 22.5) && (wf < 67.5)) Deg = "NE";
                      if ((wf >= 67.5) && (wf < 112.5)) Deg = "E";
                      if ((wf >= 112.5) && (wf < 157.5)) Deg = "SE";
                      if ((wf >= 157.5) && (wf < 202.5)) Deg = "S";
                      if ((wf >= 202.5) && (wf < 247.5)) Deg = "SW";
                      if ((wf >= 247.5) && (wf < 292.5)) Deg = "W";
                      if ((wf >= 292.5) && (wf < 337.5)) Deg = "NW";
                      if (wf >= 337.5) Deg = "N";
                    }
                }
            }
            else                          // It's a String
            {
                index = val1 + tags[tags_i].length() + 1;
                val1 = tag.indexOf("\"", index) + 1;
                index = val1 + 2;
                val2 = tag.indexOf("\"", index);
                tempx = tag.substring(val1, val2);
                if (debug) Serial.print("            ");
                if (debug) Serial.println(tempx);
                if (tags[tags_i] == "description")
                {
                    ret_string += tempx;
                    Desc = tempx;
                }
                if (tags[tags_i] == "name")
                {
                    ret_string += ",";
                    ret_string += tempx;
                    Name = tempx;
                    Serial.println(Name);
                }
            }
            tags_i += 1;
        }
    }
    if (debug) 
    {
      Serial.println(ret_string); 
      Serial.println(Speed);
      Serial.println(Deg);
    }
    return ret_string;
    
}
